use chrono;
use std::collections::HashMap;
use std::env;
use std::ffi::OsString;
use std::fs;
use std::fs::File;
use std::io::{self, BufRead};
use std::path::{Path, PathBuf};
use std::process::exit;
//use std::process::Command;
use html_editor::operation::*;
use html_editor::{parse, Doctype, Node};
use regex::Regex;

#[cfg(unix)]
use nix::{
    sys::wait::wait,
    unistd::{
        fork,
        ForkResult::{Child, Parent},
    },
};

#[cfg(unix)]
use std::thread::available_parallelism;

// ChangeLog
// 0.1.0: initial release
// 0.1.1: add cmark as formatter
// 0.1.2: get version from Cargo.toml
//        rearrange code
//        add "tidy" option in settings
// 0.1.3: put <br> inside <a> in left pane

const LEFTQ: &str = "YYY";
const LEFTQQ: &str = "yyy";
const RIGHTQ: &str = "ZZZ";
const RIGHTQQ: &str = "zzz";
const VERSION: &str = env!("CARGO_PKG_VERSION");

struct Constants {
    // things not changed by handle_dir()
    dirs: Vec<OsString>, // directory names, from the directory where wdcms is started to startdir
    maxfork: usize,      // max number of parallel processes
    mode: OsString,      // command line parameter
    outputdir: OsString, // output to this directory: root of website
    req: regex::Regex,   // regex for hendling YYY...ZZZ
    reqq: regex::Regex,  // regex for handling yyy...zzz
    sourceroot: OsString, // name of the direcory that contains wdcms.root and template.html
    startdir: OsString,  // name of directory where wdcms is called
    copyright: String,
}

impl Constants {
    fn new() -> Constants {
        Constants {
            dirs: Vec::new(),
            maxfork: 0,
            mode: OsString::new(),
            outputdir: OsString::new(),
            req: Regex::new(&(LEFTQ.to_string() + "(.*?)" + RIGHTQ)).unwrap(),
            reqq: Regex::new(&(LEFTQQ.to_string() + "(.*?)" + RIGHTQQ)).unwrap(),
            sourceroot: OsString::new(),
            startdir: OsString::new(),

            copyright: format!(
                "<!-- Generated by wdcms/rust version {} Copyright 2021, 2022 W. Vermin -->",
                version()
                ),
        }
    }
}

struct Counters {
    forked: usize,
    nummd: usize,
    numdirs: usize,
    numcss: usize,
    numjs: usize,
    numcopy: usize,
    numtrans: usize,
    numsettings: usize,
    nummenu: usize,
    numhead: usize,
}
impl Counters {
    fn new() -> Counters {
        Counters {
            forked: 0,
            nummd: 0,
            numdirs: 0,
            numcss: 0,
            numjs: 0,
            numcopy: 0,
            numtrans: 0,
            numsettings: 0,
            nummenu: 0,
            numhead: 0,
        }
    }
}

#[derive(Clone)]
struct Page {
    // for a page, things that change during traversing by handle_dir()
    content: String,                   // content, in general a markdown page
    css: Vec<(usize, OsString)>,       // list of pevious .css files
    cwd: OsString,                     // current directory in input
    head: Vec<String>,                 // list of addendums for <head>
    js: Vec<(usize, OsString)>,        // list of previous .js files
    level: usize,                      // recursion level
    menu: Vec<(String, String)>,       // menu
    multi: bool,                       // multiprocessing
    settings: HashMap<String, String>, // settings
    template: String,                  // template
    trans: HashMap<String, String>,    // translations
    urldir: String,                    // current directory in web
}

impl Page {
    fn new() -> Page {
        Page {
            content: "".to_string(),
            css: Vec::new(),
            head: vec!["".to_string(); 0],
            cwd: OsString::new(),
            js: Vec::new(),
            level: 0,
            menu: Vec::new(),
            multi: false,
            settings: HashMap::new(),
            template: "".to_string(),
            trans: HashMap::new(),
            urldir: "".to_string(),
        }
    }
}

fn abspath(p: &OsString) -> OsString {
    let s = PathBuf::from(p);
    match fs::canonicalize(&s) {
        Ok(x) => {
            return OsString::from(x);
        }
        Err(_) => {
            return OsString::new();
        }
    }
}

fn basename(p: &OsString) -> OsString {
    if let Some(x) = PathBuf::from(p).file_name() {
        x.to_os_string()
    } else {
        OsString::new()
    }
}

fn changeattr(attrs: Vec<(String, String)>, to_base: &OsString) -> Vec<(String, String)> {
    let mut out = Vec::new();
    for item in attrs {
        let mut x = item.clone();
        if item.0 == "href" || item.0 == "src" {
            let c = x.1.to_string();
            if c.len() >= 1 && &c[0..1] == "/" {
                if c.len() >= 2 && &c[1..2] == "/" {
                    x.1 = c[1..].to_string();
                } else {
                    x.1 = os_to_string(&to_base) + &c;
                }
            }
        }
        out.push(x);
    }
    return out;
}

fn chdir(dir: &OsString) {
    match std::env::set_current_dir(&dir) {
        Ok(_) => {}
        Err(_) => {
            panic!("cannot chdir to {:?}", dir);
        }
    }
}

fn copyfile(from: &OsString, to: &OsString) {
    match fs::copy(from, to) {
        Ok(_) => {}
        Err(_) => {
            panic!("Cannot copy {:?} to {:?}", from, to);
        }
    }
}

// https://stackoverflow.com/questions/26958489/how-to-copy-a-folder-recursively-in-rust
// copies contents of directory 'from' to 'to'. if 'to' is non-existant, it is created
// equivalent:
//   mkdir -p 'to'
//   cp -r 'from'/* 'to'
fn copytree<U: AsRef<Path>, V: AsRef<Path>>(from: U, to: V) -> Result<(), std::io::Error> {
    let mut stack = Vec::new();
    stack.push(PathBuf::from(from.as_ref()));

    let output_root = PathBuf::from(to.as_ref());
    let input_root = PathBuf::from(from.as_ref()).components().count();

    while let Some(working_path) = stack.pop() {
        // Generate a relative path
        let src: PathBuf = working_path.components().skip(input_root).collect();

        // Create a destination if missing
        let dest = if src.components().count() == 0 {
            output_root.clone()
        } else {
            output_root.join(&src)
        };
        if fs::metadata(&dest).is_err() {
            fs::create_dir_all(&dest)?;
        }

        for entry in fs::read_dir(working_path)? {
            let entry = entry?;
            let path = entry.path();
            if path.is_dir() {
                stack.push(path);
            } else {
                match path.file_name() {
                    Some(filename) => {
                        let dest_path = dest.join(filename);
                        copyfile(&OsString::from(path), &OsString::from(dest_path));
                    }
                    None => {
                        println!("failed: {:?}", path);
                    }
                }
            }
        }
    }

    Ok(())
} // end of copytree

fn do_format(mut soup: Vec<Node>, page: &Page, constants: &Constants) -> Vec<Node> {
    let mut formatter = page.settings["format"].clone();
    let mut data = page.content.clone();
    let (a, b) = mysplit2(data.clone(), "\n");
    if b.len() > 0 {
        let (k, v) = mysplit(a);
        if k == "format" {
            formatter = v;
            data = b;
        }
    }
    let x;
    let mut good = true;
    let html = {
        if formatter == "pandoc" {
            let out = subprocess::Exec::cmd("pandoc")
                .args(&["-t", "html", "-f", "markdown-smart-auto_identifiers"])
                .stdin(data.as_ref())
                .stdout(subprocess::Redirection::Pipe)
                .capture();

            x = out.unwrap().stdout;
            String::from_utf8_lossy(&x).to_string()
        } else if formatter == "mediawiki" {
            let out = subprocess::Exec::cmd("pandoc")
                .args(&["-t", "html", "-f", "mediawiki-smart-auto_identifiers"])
                .stdin(data.as_ref())
                .stdout(subprocess::Redirection::Pipe)
                .capture();

            x = out.unwrap().stdout;
            String::from_utf8_lossy(&x).to_string()
        } else if formatter == "cmark" {
            let out = subprocess::Exec::cmd("cmark")
                .args(&["--unsafe", "-t", "html"])
                .stdin(data.as_ref())
                .stdout(subprocess::Redirection::Pipe)
                .capture();

            x = out.unwrap().stdout;
            String::from_utf8_lossy(&x).to_string()
        } else if formatter == "asis" {
            data
        } else {
            println!("Unknown formatter defined in ./{}: '{}'. Valid formatters are 'pandoc' 'cmark' 'asis' 'mediawiki'.", 
                     page.urldir,formatter);
            good = false;
            String::new()
        }
    };
    if !good {
        exit(1);
    }

    let s = "<div class=\"markdown\">\n".to_string() + &html + "\n</div>\n" + &constants.copyright;
    // html_editor 0.5.1 does not parse things like
    //"<div\n></div>"
    // It really needs a space in stead of a single \n
    //let s = s.replace("\n"," \n");
    // issue solved in html_editor 0.5.2
    let htmlcontent;
    match parse(&s) {
        Ok(x) => {
            htmlcontent = x;
        }
        Err(e) => {
            panic!("{} in {:?}", e, page.urldir);
        }
    };

    let selector = &Selector::from("div#content");
    for node in htmlcontent {
        soup.insert_to(selector, node);
    }

    // perform translations

    let mut txt = soup.html();
    for (k, v) in &page.trans {
        txt = txt.replace(k, v);
    }

    let re = &constants.req;
    txt = re.replace_all(&txt, "$1").to_string();

    let re = &constants.reqq;
    txt = re
        .replace_all(&txt, LEFTQ.to_string() + "${1}" + RIGHTQ)
        .to_string();

    soup = parse(&txt).unwrap();

    let to_base = level2dir(&page.level);
    soup = walkattr(&soup, &changeattr, &to_base);
    return soup;
} // end of do_format

fn esclinkname(l: &String) -> String {
    rep_(&(LEFTQ.to_string() + l + RIGHTQ))
}

fn findstart(page: Page, constants: &mut Constants) -> Result<Page, Vec<String>> {
    let mut found: bool;
    let mut dirs = Vec::<OsString>::new();
    let mut texts = Vec::new();
    texts.push(
        "Searching for root of website (containing 'wdcms.root' and 'template.html)':".to_string(),
        );
    let mut prevdir = OsString::new();
    let mut depth: usize = 0;
    loop {
        depth += 1;
        let thisdir = getcwd();
        texts.push(format!(" examining: {}", os_to_string(&thisdir)));
        dirs.insert(0, basename(&thisdir));
        found = is_readable_file("wdcms.root") && is_readable_file("template.html");
        if found || thisdir == prevdir || depth > 1000 {
            break;
        }
        prevdir = thisdir;
        chdir(&OsString::from("../"));
    }
    if found {
        constants.sourceroot = basename(&getcwd());
        constants.dirs = dirs;
        return Ok(page);
    } else {
        return Err(texts);
    }
} // end of findstart

fn getcwd() -> OsString {
    match env::current_dir() {
        Ok(x) => x.into_os_string(),
        Err(_) => panic!("Cannot get value of current directory"),
    }
}

fn getdate() -> String {
    if true {
        let date_time = chrono::offset::Utc::now();
        return format!("UTC {}", date_time.format("%Y-%m-%d %H:%M:%S"));
    }
    else{
        // for reproducible generation of website
        return "date".to_string();
    }
}

fn getdirlist(p: &OsString) -> Vec<OsString> {
    let mut l = Vec::new();
    let paths = fs::read_dir(p).unwrap();
    for path in paths {
        let p = path.unwrap().path().into_os_string();
        let mut p = os_to_string(&p);
        if p.starts_with("./") {
            p = p[2..].to_string();
        }
        l.push(OsString::from(p));
    }
    return l;
}


fn handle_dir(mut page: Page, constants: &Constants, counters: &mut Counters) {
    counters.numdirs += 1;
    let urldir = OsString::from(&page.urldir);
    let level = page.level;
    page.cwd = getcwd();
    let mut d = constants.outputdir.clone();
    d.push(OsString::from("/"));
    d.push(&urldir);
    match fs::create_dir_all(&d) {
        Ok(_) => {}
        Err(_) => {
            panic!("Cannot create dir: {:?}", d);
        }
    };
    let mut ls = getdirlist(&OsString::from("."));
    ls.sort();

    let mut visit = constants.sourceroot.clone();
    visit.push(&urldir);
    println!("Visiting: {:?}", visit);

    let f = OsString::from("wdcms.copy");
    if is_readable_dir(&f) {
        let lsc = getdirlist(&f);
        for fc in lsc {
            if is_readable_file(&fc) {
                counters.numcopy += 1;
                let mut q = constants.outputdir.clone();
                q.push("/");
                if urldir.len() > 0 {
                    q.push(&urldir);
                    q.push("/");
                }
                q.push(basename(&fc));
                copyfile(&fc, &q);
            } else if is_readable_dir(&fc) {
                counters.numcopy += 1;
                let mut q = constants.outputdir.clone();
                q.push("/");
                if urldir.len() > 0 {
                    q.push(&urldir);
                    q.push("/");
                }
                q.push(basename(&fc));
                match copytree(&fc, &q) {
                    Ok(_) => {}
                    Err(_) => {
                        panic!("Cannot copytree {:?} to {:?}", fc, q);
                    }
                };
            }
        }
    }
    let f = OsString::from("template.html");
    if let Ok(x) = fs::read_to_string(f) {
        page.template = x;
    }

    let f = OsString::from("head.html");
    if let Ok(x) = fs::read_to_string(f) {
        counters.numhead += 1;
        page.head.push(x);
    }

    let f = OsString::from("content.txt");
    if let Ok(x) = fs::read_to_string(f) {
        page.content = x;
    }

    let f = OsString::from("trans.txt");
    {
        let v = read_into_vec(&f);
        if v.len() > 0 {
            page = handle_translations(page, &v);
            counters.numtrans += 1;
        }
    }

    let f = OsString::from("settings.txt");
    {
        let v = read_into_vec(&f);
        if v.len() > 0 {
            page = handle_settings(page, &v);
            counters.numsettings += 1;
        }
    }

    let f = OsString::from("menu.txt");
    {
        let v = read_into_vec(&f);
        if v.len() > 0 {
            page = handle_menu(page, &v);
            counters.nummenu += 1;
        }
    }
    let mut ls1 = ls.clone();

    for f in ls {
        let s = os_to_string(&f);
        if s.ends_with(".css") || s.ends_with(".js") {
            let mut from = constants.startdir.clone();
            if urldir.len() > 0 {
                from.push("/");
                from.push(&urldir);
            }
            from.push("/");
            from.push(basename(&f));
            let mut to = constants.outputdir.clone();
            if urldir.len() > 0 {
                to.push("/");
                to.push(&urldir);
            }
            to.push("/");
            to.push(basename(&f));
            copyfile(&from, &to);
            if s.ends_with(".css") {
                counters.numcss += 1;
                page.css.push((level, f));
            } else {
                counters.numjs += 1;
                page.js.push((level, f));
            }
        }
    }

    if constants.maxfork < 2 {
        page.multi = false;
        handle_template(&page, constants);

        counters.nummd += 1;
    } else {
        #[cfg(unix)]
        {
            page.multi = true;
            loop {
                if counters.forked < constants.maxfork {
                    let pid = unsafe { fork() };
                    match pid.expect("Fork failed") {
                        Child => {
                            handle_template(&page, constants);
                            exit(0);
                        }
                        Parent { child: _ } => {
                            counters.forked += 1;
                            counters.nummd += 1;
                            break;
                        }
                    }
                } else {
                    match wait() {
                        Ok(_) => counters.forked -= 1,
                        Err(e) => panic!("Invalid wait call: {}", e),
                    }
                }
            }
        }
    }

    if constants.mode.len() > 0 {
        if constants.mode == "." && page.level >= constants.dirs.len() - 1 {
            return;
        }
        if page.level < constants.dirs.len() - 1 {
            ls1 = vec![constants.dirs[level + 1].clone()];
        }
    }
    for f in ls1 {
        if !is_symlink(&f) && is_readable_dir(&f) && f != "wdcms.copy" {
            let curdir = getcwd();
            chdir(&f);
            page.level += 1;
            page.urldir = os_to_string(&urldir) + "/" + &os_to_string(&f);
            page.cwd = getcwd();
            handle_dir(page.clone(), constants, counters);
            chdir(&curdir);
            page.cwd = curdir;
            page.level -= 1;
        }
    }
} // end of handle_dir

fn handle_menu(mut page: Page, lines: &Vec<String>) -> Page {
    let mut firstmenu = true;
    for line in lines {
        if !line.starts_with("#") {
            let (k, v) = mysplit(&line);
            if v.len() > 0 {
                if k == "menu" {
                    if firstmenu {
                        firstmenu = false;
                        page.menu = Vec::new();
                    }
                    let (l, w) = mysplit(&v);
                    if l.len() > 0 {
                        page.menu.push((l, w));
                    }
                }
            }
        }
    }
    return page;
} // end of handle_menu

fn handle_settings(mut page: Page, lines: &Vec<String>) -> Page {
    for line in lines {
        if !line.starts_with("#") {
            let (k, v) = mysplit(&line);
            if v.len() > 0 {
                let rc = match k.as_str() {
                    "format" | "superfish" | "tidy" => true,
                    _ => false,
                };
                if rc {
                    page.settings.insert(k, v);
                } else {
                    panic!("Invalid settings in {}, quitting", k); // todo: urldir
                }
            }
        }
    }
    return page;
}

fn handle_superfish(cwd: OsString, level: &usize, name: &String) -> Node {
    let mut d = cwd;
    d.push("/");
    d.push(level2dir(&level));
    d.push("/");
    d.push(name);
    let mut ls = getdirlist(&d);
    ls.sort();
    let mut lis = Vec::new();
    for f in ls {
        let sf = os_to_string(&f);
        let bs = os_to_string(&basename(&f));
        if is_readable_dir(&f) && !sf.ends_with("wdcms.copy") {
            let href = name.clone() + "/" + &bs + "/index.html";
            let a: Node = Node::new_element(
                "a",
                vec![("href", &href), ("class", "sf-dropdown")],
                vec![Node::Text(esclinkname(&bs))],
                );
            let li: Node = Node::new_element("li", vec![("class", "menu-item")], vec![(a)]);
            lis.push(newline());
            lis.push(li);
        }
    }
    let ul: Node = Node::new_element("ul", vec![], lis);

    return ul;
} // end of handle_superfish

fn handle_template(page: &Page, constants: &Constants) {
    let css = &page.css;
    let head = &page.head;
    let js = &page.js;
    let level = &page.level;
    let menu = &page.menu;
    let settings = &page.settings;
    let template = &page.template;
    let urldir = &page.urldir;

    let mut soup = parse(template).unwrap();
    let mut lis = Vec::new();
    let mut first = true;
    for item in menu {
        let mut href = item.0.clone();
        href = href + "/index.html";
        let name: String;
        let mut wdcms_name = item.0.clone();
        if item.0 == "/" {
            name = "home".to_string();
            href = "/index.html".to_string();
            wdcms_name = "/.".to_string();
        } else {
            name = os_to_string(&basename(&OsString::from(&item.0)));
        }
        let name = esclinkname(&name);

        // add specific class to main menu anchors
        let mut class = "link".to_string();
        if first {
            let mut x = urldir.clone();
            if x.len() == 0 {
                x = "/.".to_string();
            } else {
                x = "/".to_string();
                x = mysplit1(x + &urldir[1..], "/").0;
            }
            if x == wdcms_name {
                class = class + " main-menu-active";
                first = false;
            }
        }

        let a: Node = Node::new_element(
            "a",
            vec![
            ("href", &href),
            ("class", &class),
            ("title", &item.1),
            ("wdcms_name", &wdcms_name),
            ],
            vec![Node::Text(name)],
            );
        lis.push(newline());
        let mut a = vec![a];
        if settings["superfish"] == "yes" {
            let s = handle_superfish(page.cwd.clone(), level, &wdcms_name);
            a.push(s);
        }
        let li: Node = Node::new_element("li", vec![("class", "menu-item")], a);

        lis.push(li);
        lis.push(newline());
    }
    lis.push(newline());
    let ul: Node = Node::new_element(
        "ul",
        vec![("id", "sf-main-menu"), ("class", "clearfix menu sf-menu")],
        lis,
        );
    soup.insert_to(&Selector::from("div#main-menu"), ul);

    let head_selector = Selector::from("head");
    // handle head
    for f in head {
        let head_dom = parse(&f).unwrap();
        for item in &head_dom {
            soup.insert_to(&head_selector, item.clone());
        }
    }

    // handle css
    for f in css {
        let r = os_to_string(&level2dir(&(level - f.0)));
        let l = os_to_string(&f.1);
        //let p = soup.new_tag("link", href=r+"/"+l)
        let link: Node = Node::new_element(
            "link",
            vec![("href", &(r + "/" + &l)), ("rel", "stylesheet")],
            vec![],
            );
        soup.insert_to(&head_selector, link);
        soup.insert_to(&head_selector, newline());
    }

    // handle js
    for f in js {
        let r = os_to_string(&level2dir(&(level - f.0)));
        let l = os_to_string(&f.1);
        let script: Node = Node::new_element("script", vec![("src", &(r + "/" + &l))], vec![]);
        soup.insert_to(&head_selector, script);
        soup.insert_to(&head_selector, newline());
    }

    // left navigation
    let navigation_selector = Selector::from("div#leftlinkstop");

    // find all directories on the same level

    let x = getdirlist(&page.cwd);
    let mut ls = Vec::new();
    for item in x{
        ls.push(basename(&item));
    }
    ls.sort();
    let mut found = false;
    let br = Node::new_element("br", vec![], vec![]);
    for dir in ls {
        let sdir = os_to_string(&dir);
        if is_readable_dir(&dir) && !sdir.ends_with("wdcms.copy") {
            found = true;
            let a = Node::new_element(
                "a",
                vec![("href", &(sdir.clone() + "/index.html"))],
                vec![Node::Text(esclinkname(&sdir)+"<br>")],
                );
            soup.insert_to(&navigation_selector, a);
            //soup.insert_to(&navigation_selector, br.clone());
            // No: put <br> inside <a>
            soup.insert_to(&navigation_selector, newline());
        }
    }
    if found {
        soup.insert_to(
            &Selector::from("div#leftlinkstop_header"),
            Node::Text(esclinkname(&os_to_string(&basename(&getcwd())))),
            );
    }

    // create links to parent's parent directories
    let selector = &Selector::from("div#leftlinksbot");
    // find all directories leading to this

    let mut dir = "./..".to_string();
    let mut found = false;
    if *level > 1 {
        for _i in 0..level - 1 {
            found = true;
            let a = Node::new_element(
                "a",
                vec![("href", &(dir.clone() + "/index.html"))],
                vec![Node::Text(esclinkname(&os_to_string(&basename(&abspath(
                                    &OsString::from(&dir),
                                    )))))],
                                    );
            soup.insert_to(&selector, a);
            soup.insert_to(&selector, br.clone());
            soup.insert_to(&selector, newline());
            dir = dir + "/..";
        }
    }
    if found {
        let selector = &Selector::from("div#leftlinksbot_header");
        soup.insert_to(&selector, Node::Text("_____".to_string()));
    }
    soup = do_format(soup, page, constants);
    write_page(soup, page, constants);
} // end of handle_template

fn handle_translations(mut page: Page, lines: &Vec<String>) -> Page {
    for line in lines {
        if !line.starts_with("#") {
            let (k, v) = mysplit(&line);
            if v.len() > 0 {
                page.trans.insert(linkname(&k), v.clone());
                let q = k.find('_');
                if let Some(_) = q {
                    let k = esclinkname(&k);
                    page.trans.insert(linkname(&k), v);
                }
            }
        }
    }
    return page;
}

fn init(mut page: Page, constants: &mut Constants, i: &OsString) -> Page {
    constants.mode = i.clone();

    let lines = vec![
        "wdcms_date: ".to_string() + &getdate(),
        "wdcms_version: ".to_string() + &version(),
        "leftq: ".to_string() + LEFTQQ,
        "rightq: ".to_string() + RIGHTQQ,
    ];
    page = handle_translations(page, &lines);

    let lines = vec![
        "superfish: yes".to_string(),
        "format: pandoc".to_string(),
        "tidy:      yes".to_string(),
    ];
    page = handle_settings(page, &lines);

    match findstart(page, constants) {
        Ok(x) => {
            page = x;
        }
        Err(x) => {
            for t in x {
                println!("{}", t);
            }
            println!("No directory containing 'wdcms.root' and 'template.html' found, quitting.");
            exit(1);
        }
    }

    return page;
} // end of init

fn is_readable_dir<P>(dirname: P) -> bool
where
P: AsRef<Path>,
{
    if let Ok(_) = std::fs::File::open(&dirname) {
        if let Ok(x) = std::fs::metadata(dirname) {
            return x.is_dir();
        }
    }
    return false;
}

fn is_readable_file<P>(filename: P) -> bool
where
P: AsRef<Path>,
{
    if let Ok(_) = std::fs::File::open(&filename) {
        if let Ok(x) = std::fs::metadata(filename) {
            return x.is_file();
        }
    }
    return false;
}

fn is_symlink(f: &OsString) -> bool {
    let metadata = fs::symlink_metadata(f).unwrap();
    let file_type = metadata.file_type();
    return file_type.is_symlink();
}

fn level2dir(level: &usize) -> OsString {
    let mut d = OsString::new();
    d.push(".");
    for _i in 0..*level {
        d.push("/..");
    }
    return d;
}

fn linkname(l: &String) -> String {
    format!("{}{}{}", LEFTQ, l, RIGHTQ)
}

fn main() {
    tell(("".to_string(), 0, 0, ("".to_string(), "".to_string())));
    let args: Vec<String> = env::args().collect();
    println!("wdcms.rs version {}", version());
    {
        let machine_kind = if cfg!(unix) {
            "unix"
        } else if cfg!(windows) {
            "windows"
        } else if cfg!(macos) {
            "macos"
        } else {
            "unknown"
        };
        println!("Detected system type:      {}", machine_kind);
        #[cfg(unix)]
        {
            println!(
                "Number of (virtual) cpu's: {}",
                available_parallelism().unwrap().get()
                );
        }
    }

    let mut page = Page::new();
    let mut counters = Counters::new();
    let mut constants = Constants::new();
    if args.len() == 1 {
        page = init(page, &mut constants, &OsString::from(""));
    } else {
        let arg = args[1].clone();
        if arg != "." && arg != "/" {
            panic!(
                "argument must be \".\" or \"/\", argument given is: \"{}\"  ",
                arg
                );
        }
        page = init(page, &mut constants, &OsString::from(args[1].clone()));
    }

    start(page, constants, &mut counters);

    println!("Generated:");
    println!("   index.html   : {}", counters.nummd);
    println!("Copied:");
    println!("   folders      : {}", counters.numdirs);
    println!("   *.css        : {}", counters.numcss);
    println!("   *.js         : {}", counters.numjs);
    println!("   copy         : {}", counters.numcopy);
    println!("Used:");
    println!("   content.txt  : {}", counters.nummd);
    println!("   trans.txt    : {}", counters.numtrans);
    println!("   settings.txt : {}", counters.numsettings);
    println!("   menu.txt     : {}", counters.nummenu);
    println!("   head.html    : {}", counters.numhead);

    #[cfg(unix)]
    {
        // wait for all forked processes to end
        // println!("processes left: {}",counters.forked);
        while counters.forked > 0 {
            match wait() {
                Ok(_) => counters.forked -= 1,
                Err(e) => panic!("Invalid wait call: {}", e),
            }
        }
    }
}
// end of main

fn mysplit<T: AsRef<str>>(s: T) -> (String, String) {
    return mysplit1(s.as_ref(), &":".to_string());
}

// discards comments:
fn mysplit1<T: AsRef<str>, U: AsRef<str>>(s: T, t: U) -> (String, String) {
    let s1 = s.as_ref();
    let s: &str;
    match s1.find('#') {
        Some(x) => {
            s = &s1[..x];
        }
        None => s = s1,
    }
    let mut a = s.splitn(2, t.as_ref());
    let first = a.next().unwrap().trim().to_string();
    let second = if let Some(x) = a.next() { x } else { "" }
    .trim()
        .to_string();
    return (first, second);
}

fn mysplit2<T: AsRef<str>, U: AsRef<str>>(s: T, t: U) -> (String, String) {
    let mut a = s.as_ref().splitn(2, t.as_ref());
    let first = a.next().unwrap().trim().to_string();
    let second = if let Some(x) = a.next() { x } else { "" }
    .trim()
        .to_string();
    return (first, second);
}

fn newline() -> Node {
    return Node::Text("\n".to_string());
}

fn os_to_string(s: &OsString) -> String {
    match s.clone().into_string() {
        Ok(x) => x,
        Err(e) => panic!("Invalid UTF-8 sequence: {:?}", e),
    }
}

fn read_into_vec(f: &OsString) -> Vec<String> {
    let mut x = Vec::new();
    if let Ok(lines) = read_lines(f) {
        for line in lines {
            if let Ok(ip) = line {
                x.push(ip);
            }
        }
        return x;
    }
    return Vec::<String>::new();
}

fn read_lines<P>(filename: P) -> io::Result<io::Lines<io::BufReader<File>>>
where
P: AsRef<Path>,
{
    let file = File::open(filename)?;
    Ok(io::BufReader::new(file).lines())
}

fn rep_(l: &String) -> String {
    l.replace("_", " ").trim().to_string()
}

fn start(mut page: Page, mut constants: Constants, counters: &mut Counters) {
    constants.startdir = getcwd();
    println!("Creating website");
    println!("Input starts in {:?}", constants.startdir);
    constants.outputdir = OsString::from("/tmp/website");
    if let Ok(lines) = read_lines("wdcms.root") {
        for line in lines {
            if let Ok(ip) = line {
                let (k, v) = mysplit1(ip, ":");
                if v.len() > 0 {
                    let mut par = 0;
                    match k.as_ref() {
                        "outputdir" => {
                            constants.outputdir = OsString::from(v);
                        }
                        "parallel" => {
                            if let Ok(x) = v.to_string().parse::<i32>() {
                                par = x;
                            } else {
                            }
                        }
                        _ => {}
                    }
                    constants.maxfork = (0 * par) as usize;
                    #[cfg(unix)]
                    {
                        if par < 0 {
                            constants.maxfork = available_parallelism().unwrap().get();
                        } else {
                            constants.maxfork = par as usize;
                        }
                    }
                }
            }
        }
    } else {
        panic!("read_line: Cannot read file wdcms.root");
    }
    println!("Global settings from 'wdcms.root':");
    println!("   outputdir: {}", os_to_string(&constants.outputdir));
    println!("   parallel : {}", constants.maxfork);

    let v = read_into_vec(&OsString::from("settings.txt"));
    page = handle_settings(page, &v);
    println!("Start settings from 'settings.txt':");
    println!("   format   : {}", page.settings["format"]);
    println!("   superfish: {}", page.settings["superfish"]);
    println!("   tidy:      {}", page.settings["tidy"]);
    handle_dir(page, &constants, counters);
} // end of start

fn tell(_a: (String, i32, i64, (String, String))) {}

fn version() -> String {
    VERSION.to_string()
}

fn walkattr(
    dom: &Vec<Node>,
    f: &dyn Fn(Vec<(String, String)>, &OsString) -> Vec<(String, String)>,
    to_base: &OsString,
    ) -> Vec<Node> {
    let mut newdom = Vec::new();
    for node in dom {
        let mut x = node.clone();
        match node {
            Node::Doctype(d) => match d {
                Doctype::Html => {}
                Doctype::Xml {
                    version: _x,
                    encoding: _y,
                } => {}
            },
            Node::Text(_s) => {}
            Node::Comment(_s) => {}
            Node::Element {
                name: n,
                attrs: a,
                children: c,
            } => {
                x = Node::Element {
                    name: n.to_string(),
                    attrs: f(a.to_vec(), to_base),
                    children: walkattr(&c, &f, to_base),
                };
            }
        }
        newdom.push(x);
    }
    return newdom;
}

fn write_page(soup: Vec<Node>, page: &Page, constants: &Constants)  {
    let mut out = OsString::from(&constants.outputdir);
    out.push(page.urldir.clone());
    out.push(OsString::from("/index.html"));
    let mut html = soup.html();
    if page.settings["tidy"] == "yes" {
        let txt = subprocess::Exec::cmd("tidy")
            .args(&["-i", "-q", "--show-warnings", "no"])
            .stdin(html.as_ref())
            .stdout(subprocess::Redirection::Pipe)
            .capture();
        let x = txt.unwrap().stdout;
        html = String::from_utf8_lossy(&x).to_string();
    }
    fs::write(out, html).expect("Unable to write file");
}
